'use strict';
(function($){

/**
 * @author Martin Zilz
 * @copyright Kreativkonzentrat GbR
 *
 * Responsive Items Slider
 * =======================
 * Some Assumptions to be made:
 * 
 * 1) Items are equal in width (set it in CSS, e.g. width:25%)
 * 2) There is no Margin on Items (script resets it automaticly)
 * 3) Contents are Styled with Border-Box Model (script sets model automaticly)
 */
window.kkItemSlider = function(options){
	this._init(options);
}

kkItemSlider.prototype = {

	_init: function(options){
		this.options = $.extend({}, {
			containerSelector: ".kk-item-slider",	// selector of the root container
			itemSelector: ".kk-slider-item",		// selector of the items inside the container
			minItemSize: 200,						// min width of items. if smaller items visible are reduced
			maxItemScroll: 10,						// maximum amount of items that are overscrolled
			scrollTime: 300,						// scroll Time in milliseconds
			showNavigation: true,					// should we see the navigation arrows?
			clickToggle: false,						// does clicking switch to the next slides?
			autoStart: false,						// should the slider automaticly rotate?
			autoStartSwitchDelay: 4000,				// delay it takes between automatic switches
			allowOverscroll: false,
			direction: "horizontal"
		}, options);

		// just for fallback reason. it's called minItemSize now
		if(typeof(options.minItemWidth) !== "undefined"){
			this.options.minItemSize = options.minItemWidth;
		}

		this._build();
	},

	_build: function(){
		var self = this;
		$(this.options.containerSelector).each(function(){
			new kkItemSliderInstance(self.options, $(this), $(this).find(self.options.itemSelector));
		});
	}
}

var kkItemSliderInstance = function(options, container, items){
	this._init(options, container, items);
}


//
// An instance of our Slider
//
kkItemSliderInstance.prototype = {
	/**
	 * Our Slider constructor
	 */
	_init: function(options, container, items){

		var self = this;
		// do not let this run twice on the same container
		if(container.hasClass('handled')){
			return;
		}

		// attach data to this
		container.data('kkItemSlider', this);

		// if our container is not visible and therefore we cannot get any
		// measures, try again all 300ms until it becomes visible!
		if(!this._isVisible(container)){
			
			// hide it until we can show it
			container.css('opacity', '0');

			setTimeout(function(){
				self._init(options, container, items);
			}, 300);

			return;
		}

		// build the DOM as we need it
		this.scrollIndex = 0;   // index of the most left visible item in the slider
		this.jContainer = container;
		this.jContainer.addClass('handled');
		this.jItems = items;	// our slider items
		this.options = options;
		this.jContainer.addClass('kk-item-slider'); 
		this.jItems.addClass('kk-slider-item');
		this.jContainer.wrap('<div class="kk-item-slider-root"></div>');    // the wrap arround our container
		this.jContainerWrap = $(this.jContainer.closest('.kk-item-slider-root'));   // the container under which the items are grouped
		this.jContainerWrap.wrap('<div class="kk-item-slider-style-wrap"></div>');
		this.jContainerStyleWrap = $(this.jContainerWrap.closest('.kk-item-slider-style-wrap'));    // the container under which the items are grouped

		// on vertical direction set a minimum height
		if(this.options.direction === "vertical"){
			this.jContainer.css('min-height', this.options.minItemSize + 'px');
		}

		// build navigation, if told to do so
		if(this.options.showNavigation === true){
			this.jPrevious = $('<div class="kk-item-slider-prev"></div>');
			this.jNext = $('<div class="kk-item-slider-next"></div>');
			this.jContainerStyleWrap.append(this.jPrevious);
			this.jContainerStyleWrap.append(this.jNext);
			this.jContainerStyleWrap.addClass('has-nav');
		}

		// add directional class
		this.jContainerStyleWrap.addClass('kk-slider-' + options.direction);

		// check if we need to have another starting index
		this.jItems.each(function(idx){
			if($(this).attr('data-start-item')){
				self.scrollIndex = idx;
				// break loop
				return false;
			}
		});

		this._setup();

		// fade in container
		this.jContainer.css('opacity', '');
	},

	/**
	 * checks if the parent container size changed since we have called
	 * that function the last time
	 */
	_checkParentSizeChanged: function() {
		// can we skip layout because the container didnt change?
		var jParent = this.jContainerStyleWrap.parent();
		var newParentWidth = jParent.innerWidth();
		if(this.oldParentWidth && this.oldParentWidth == newParentWidth){
			return false;
		}
		this.oldParentWidth = newParentWidth;
		return true;
	},
	
	/**
	 * Initialize the Slider a first time - bind Events
	 */
	_setup: function(){
		var self = this;

		// layout ourselfs when a parent has changed its layout
		self.jContainer.on('drop.parent.resize', function(parent){
			if(!self._inLayoutDelayPhase && self._checkParentSizeChanged()){
				self._layout();
			}
		});

		// handle window width resize (relayout)
		$(window).resize(function(){
			if(!self._inLayoutDelayPhase && self._checkParentSizeChanged()){
				self._layout();
			}
		});

		// handle clicktoggle
		if(this.options.clickToggle === true){
			this.jContainerStyleWrap.click(function(){
				self._autostartStop = true;
				self.next();
			});
		}

		// handle navigation
		if(this.options.showNavigation === true){
			this.jNext.click(function(e){
				self._autostartStop = true;
				e.preventDefault();
				e.stopPropagation();
				self.next();
			});

			this.jPrevious.click(function(e){
				self._autostartStop = true;
				e.preventDefault();
				e.stopPropagation();
				self.previous();
			});
		}

		// handle arrow keys for navigation
		$('body').bind('keydown', function(e){
			// check if we are very close to the center of the screen
			var pos = self.jContainerStyleWrap.offset().top + self.jContainerStyleWrap.height()/2 - $(document).scrollTop();
			var windowHeight = $(window).height();
			var relation = (pos - windowHeight/2)/windowHeight;

			if(Math.abs(relation) < 0.2 && !$(':focus').length){
				switch(e.which){
					case 37:
					self.previous();
					break;
					case 39:
					self.next();
					break;
				}
			}

		});
		
		// handle touch events
		this.jContainer.bind('touchstart', function(e){
			self._dragStart({
					x: e.originalEvent.touches[0].clientX,
					y: e.originalEvent.touches[0].clientY
				}, e);
		});
		$('body').bind('touchmove', function(e){
			self._dragMove({
					x: e.originalEvent.touches[0].clientX,
					y: e.originalEvent.touches[0].clientY
				}, e);
		});
		$('body').bind('touchend', function(e){
			self._dragEnd(e);
		});

		// handle image loads inside the items. we need to relayout eventually if no width or height was set for the image
		this.jItems.find('img').each(function(){
			$(this).load(function(){
				if(self._loadTimeoutHandle){
					clearTimeout(self._loadTimeoutHandle);
				}
				self._loadTimeoutHandle = setTimeout(function(){
					self._layout();
				}, 300);
			});
		});

		// autoStart?
		if(this.options.autoStart === true){
			clearInterval(this._autostartInterval);
			this._autostartInterval = setInterval(function(){
				if(self._autostartStop) return clearInterval(self._autostartInterval);
				self.next();
			}, this.options.autoStartSwitchDelay);
		}

		// initial layout
		self._layout();
	},
	/**
	 * Relayouts the slider whenever needed. Calculates the important stuff
	 */
	_layout: function(){
		var self = this;
		// it might happen that a parent is set to display none. In this case we delay the layout phase
		// until we get visible again
		if(!this._isVisible(this.jContainerWrap)){
			this._inLayoutDelayPhase = true;
			if(this._timeoutHandle){
				clearTimeout(this._timeoutHandle);
			}
			this._timeoutHandle = setTimeout(function(){
				self._layout();
			}, 300);
			return;
		}

		this._inLayoutDelayPhase = false;

		// reset everything so that we can measure correctly
		this.jContainer.removeClass('initialized');

		this.jContainerWrap.css({
			width: '',
			height: ''
		});

		// reset styles on the items
		this.jItems.css({
			'width': '',
			'position': '',
			'left': '',
			'top' : '',
			'height': ''
		});


		// reset offset
		this.jContainer.css({
			left: ''
		});

		// reset styles on the wrapper
		this.jContainerStyleWrap
			.css('width', '')
			.css('height', '')
			.css('marginTop', '')
			.css('marginLeft', '')
			.css('marginRight', '')
			.css('marginBottom', '')
			.css('paddingTop', '')
			.css('paddingLeft', '')
			.css('paddingRight', '')
			.css('paddingBottom', '');

		// calculate new base values
		if(!this.jContainer.hasClass('initialized')){
			if(this.options.direction === "horizontal"){
				this.jContainerStyleWrap.css('width', this.jContainer.outerWidth() + 'px')
			}else{
				this.jContainerStyleWrap.css('height', this.jContainer.outerHeight() + 'px')
			}
			this.jContainerStyleWrap
				.css('marginTop', this.jContainer.css('marginTop'))
				.css('marginLeft', this.jContainer.css('marginLeft'))
				.css('marginRight', this.jContainer.css('marginRight'))
				.css('marginBottom', this.jContainer.css('marginBottom'))
				.css('paddingTop', this.jContainer.css('paddingTop'))
				.css('paddingLeft', this.jContainer.css('paddingLeft'))
				.css('paddingRight', this.jContainer.css('paddingRight'))
				.css('paddingBottom', this.jContainer.css('paddingBottom'));
		}


		// recalc the with now and set us to initialized state
		this.containerSize = (this.options.direction === "horizontal" ? this.jContainerWrap.outerWidth() : this.jContainerWrap.outerHeight());

		// calculate the width and visible number of items
		var numItemsFit = this._checkHowManyItemsFit(),
			size;

		size = this.containerSize / numItemsFit;

		// bigger than minimum width
		if(size >= this.options.minItemSize){
			this.itemsInView = numItemsFit;
			this.itemSize = size;
		}
		// only one item and smaller than minimum width
		else if(numItemsFit <= 1){
			this.itemsInView = 1;
			this.itemSize = this.containerSize;
		}
		// more than one item, but small than minimum width
		else{
			numItemsFit = Math.floor(this.containerSize / this.options.minItemSize);
			this.itemSize = this.containerSize / numItemsFit;
			this.itemsInView = numItemsFit;
		}

		// if there are all items visible set a class for this
		if(numItemsFit >= this.jItems.length){
			this.jContainerStyleWrap.addClass('all-visible');
		}else{
			this.jContainerStyleWrap.removeClass('all-visible');
		}

		this.jContainer.addClass('initialized');

		// set new item size
		var maxSize = 0;
		for(var i=0; i<this.jItems.length; i++){
			var item = $(this.jItems[i]);

			if(this.options.direction === "horizontal"){
				item.css({
					'width': this.itemSize + 'px',
					'position': 'absolute',
					'left': (i * this.itemSize) + 'px',
					'top': 0
				});
				maxSize = Math.max(maxSize, item.outerHeight());
			}else{
				item.css({
					'height': this.itemSize + 'px',
					'position': 'absolute',
					'top': (i * this.itemSize) + 'px',
					'left': 0
				});
				maxSize = Math.max(maxSize, item.outerWidth());
			}
		}

		// set container height and items height, based on the highest item
		if(this.options.direction === "horizontal"){
			this.jItems.css('height', maxSize + 'px');
			this.jContainerWrap.css('height', maxSize + 'px');
		}
		else{
			this.jItems.css('width', maxSize + 'px');
			this.jContainerWrap.css('width', maxSize + 'px');
		}

		// this.containerSize = maxSize;

		// if we make the window larger, make sure that there can be no gap to the right
		if(this.scrollIndex + this.itemsInView > this.jItems.length){
			this.scrollIndex = this.jItems.length - this.itemsInView;
		}

		// rebase the offset
		this.scrollTo(this.scrollIndex, 0);
		this._checkDirectionsAllowed();
	},

	/**
	 * Scrolls the slider to some index
	 */
	scrollTo: function(index, time){
		var self = this;
		var oldScrollIndex = this.scrollIndex,
			moveClass;

		this.scrollIndex = this.jItems ? (index % (this.jItems.length - (this.itemsInView - 1))) : 0;

		// determine the scrolling direction
		if(oldScrollIndex < this.scrollIndex){
			moveClass = "kk-forwards";
		}else if(oldScrollIndex > this.scrollIndex){
			moveClass = "kk-backwards";
		}

		this._checkDirectionsAllowed();

		// update classes
		this.jContainer.addClass(moveClass);

		if(this.jItems && this.jItems.length > 0){
			this.jItems.removeClass('kk-first-in-view kk-last-in-view');
			if(this.jItems[index]) { $(this.jItems[index]).addClass('kk-first-in-view'); }
			if(this.jItems[index + this.itemsInView - 1]) { $(this.jItems[index + this.itemsInView - 1]).addClass('kk-last-in-view'); }
		}

		// do the scrolling animation
		this._moveSlidingContainer((-1.0 * parseFloat(this.scrollIndex) * parseFloat(this.itemSize)), time, function(){
			self.jContainer.removeClass('kk-forwards kk-backwards');
		});
	},

	/**
	 * Checks for CSS transition support
	 */
	_supportsTransitions: function(){
		var s = document.createElement('p').style,
		v = ['ms','O','Moz','Webkit'];
		
		if( s['transition'] == '' ) return true;
		while( v.length ){
			if(v.pop() + 'Transition' in s){
				return true;
			}
		}
		return false;
	},
	
	/**
	 * Checks for CSS transform support
	 */
	_supportsTransform: function(){
		var prefixes = 'transform WebkitTransform MozTransform OTransform msTransform'.split(' ');
		var div = document.createElement('div');
		for(var i = 0; i < prefixes.length; i++){
			if(div && div.style[prefixes[i]] !== undefined){
				return true;
			}
		}
		return false;
	},
	
	/**
	 * Moves the sliding container in the given time to the given offset using
	 * automaticly found transition method or animation
	 */
	_moveSlidingContainer: function(to, time, cb){

		if(typeof(this._fastAnimate) === 'undefined'){
			this._fastAnimate = (this._supportsTransitions() && this._supportsTransform());
		}
		// fast animation path
		if(this._fastAnimate){
			var translate = 'translate' + (this.options.direction === 'vertical' ? 'Y' : 'X') + '(' + to + 'px)';
			var css = {
				'-moz-transition' : '-moz-transform ' + time + 'ms ease-out',
				'-ms-transition' : '-ms-transform ' + time + 'ms ease-out',
				'-webkit-transition' : '-webkit-transform ' + time + 'ms ease-out',
				'-o-transition' : '-o-transform ' + time + 'ms ease-out',
				'transition' : 'transform ' + time + 'ms ease-out',
				'-moz-transform' : translate,
				'-ms-transform' : translate,
				'-webkit-transform' : translate,
				'-o-transform' : translate,
				'transform' : translate
			}

			this.jContainer.css(css);

			if(cb){
				setTimeout(function(){
					cb();
				}, time);
			}
		}
		// slow animation path
		else {
			var prop = this.options.direction === 'vertical' ? {'top' : to + 'px'} : {'left' : to + 'px'};

			this.jContainer.stop().animate(prop, time, function(){
				if(cb) cb();
			});
		}

		// buffer the offset we have set
		this._sliderOffset = to;
		
	},

	/**
	 * Scrolls the slider to the next stage.
	 * Steps can be fixed to a maximum in the settings object
	 */
	next: function(){
		var scrollOffset = this.options.maxItemScroll;
		if(this.options.maxItemScroll > this.itemsInView){
			scrollOffset = this.itemsInView;
		}
		// this is the wished new index
		var newIndex = this.scrollIndex + scrollOffset;

		// we have reached the end excactly... next step is back to zero
		if(this.scrollIndex === this.jItems.length - this.itemsInView && this.options.allowOverscroll === true){
			newIndex = 0;
		}
		// we have not reached the end completely, scroll till the end first
		else if(newIndex > this.jItems.length - this.itemsInView){
			newIndex = this.jItems.length - this.itemsInView;
		}

		this.scrollTo(newIndex, this.options.scrollTime);
	},

	/**
	 * Scrolls the slider to the previous stage.
	 * Steps can be fixed to a maximum in the settings object
	 */
	previous: function(){
		var scrollOffset = Math.min(this.options.maxItemScroll, this.itemsInView);

		// this is the wished new index
		var newIndex = this.scrollIndex - scrollOffset;

		if(newIndex < 0 && newIndex + scrollOffset > 0){
			newIndex=0;
		}
		// we have reached the end excactly... next step is back to the end
		else if(newIndex + scrollOffset === 0){
			newIndex = this.jItems.length - this.itemsInView;
		}
		// do not break when overscroll is disallowed
		if(this.options.allowOverscroll === false && this.scrollIndex - scrollOffset < 0){
			newIndex = 0;
		}

		this.scrollTo(newIndex, this.options.scrollTime);
	},

	/**
	 * Checks if the Previous / Next Buttons can be used
	 */
	_checkDirectionsAllowed: function(){
		var nextAllowed = (this.options.allowOverscroll === true) || ((this.jItems.length - (this.scrollIndex + this.itemsInView) > 0) && this.jItems.length > this.itemsInView);
		var prevAllowed = (this.options.allowOverscroll === true) || (this.scrollIndex > 0 && this.jItems.length > this.itemsInView);

		if(this.options.showNavigation === true){
			this.jNext.removeClass('disallowed');
			this.jPrevious.removeClass('disallowed');
			
			if(!nextAllowed){
				this.jNext.addClass('disallowed');
			}

			if(!prevAllowed){
				this.jPrevious.addClass('disallowed');
			}
		}
	},

	/**
	 * Checks how many items fit the available space
	 */
	_checkHowManyItemsFit: function(){
		var itemAddedSize = 0;
		var numItemsFit = 0;

		for(var i=0; i<this.jItems.length; i++){
			var itemSize = this.options.direction === "horizontal" ? 
				Math.max($(this.jItems[i]).outerWidth(), this.options.minItemSize) : 
				Math.max($(this.jItems[i]).outerHeight(), this.options.minItemSize);

			if((itemAddedSize + itemSize) > this.containerSize) {
				return (numItemsFit > 0 ? numItemsFit : 1);
			}
			else{
				itemAddedSize += itemSize;
				numItemsFit ++;
			}
		}
		return numItemsFit;
	},
	/**
	 * Checks if the slider is visible inside the dom by checking up all children
	 * for visibility
	 */
	_isVisible: function(element){
		if(!element) return false;
		return $(element).is(':hidden') ? false : true;
	},
	/**
	 * Callback when a user starts dragging
	 */
	_dragStart: function(pos, e){
		this._isDragging = true;
		this._dragStartPos = $.extend({}, pos),
		this._dragStartSliderPos = this._sliderOffset;
		this._dragDist = false;
	},
	/**
	 * Callback when a user moves the pointer after starting dragging
	 */
	_dragMove: function(pos, e){

		if(!this._isDragging) return;
		
		var pane;
		var self = this;
		var dist = {
			x: pos.x - this._dragStartPos.x,
			y: pos.y - this._dragStartPos.y,
		};
		
		switch(this.options.direction) {
			case 'horizontal' :
				pane = 'x';
			break;
			
			case 'vertical':
				pane = 'y';
			break;
		}

		// lock movement on axis
		if( ( pane === 'x' && Math.abs(dist.x) >= Math.abs(dist.y) ) || ( pane === 'y' && Math.abs(dist.y) >= Math.abs(dist.x) ) ){ 
			// lock on axis
			e.preventDefault(); e.stopPropagation();
		} else {
			// do nothing because this movement is on the wrong axis
			return;
		}
		
		self._moveSlidingContainer(this._dragStartSliderPos + dist[pane], 0);
		self._dragDist = dist;
	},
	/**
	 * Callback when a user ends dragging
	 */
	_dragEnd: function(e){

		if(!this._isDragging || !this.itemSize || !this._dragDist) return;

		this._isDragging = false;

		var dist;
		
		switch(this.options.direction) {
			case 'horizontal' :
				dist = this._dragDist.x;
			break;
			
			case 'vertical':
				dist = this._dragDist.y;
			break;
		}

		var draggedItems = -1 * (Math.sign(dist) > 0 ? Math.floor(dist / this.itemSize) : Math.ceil(dist / this.itemSize));
		
		if(Math.abs(dist % this.itemSize) > 20) draggedItems += (dist > 0 ? -1 : 1);

		var newIndex = this.scrollIndex + draggedItems;

		// keep it in bounds
		if(newIndex < 0) newIndex = 0;
		if(newIndex > this.jItems.length - this.itemsInView) newIndex = this.jItems.length - this.itemsInView;

		this.scrollTo(newIndex, 200 /*this.options.scrollTime*/);
	}
};

})(jQuery);