/**
 * Core Service
 * Provides access to otherwise global Data which is prepopulated
 * by the administration Backend
 */
angular.module('app').service('core', ['$http', '$controller', '$q', 'message', function($http, $controller, $q, message) {
	
	if (window.dropperCore) return window.dropperCore;
	
	var core = {

		settings: {},

		fn: {

			_listeners: {},

			on: function(evt, cb) {
				if (!core.fn._listeners) {
					core.fn._listeners = {};	
				}
				
				if (!core.fn._listeners[evt]) {
					core.fn._listeners[evt] = [];
				}
				
				// dont allow double registration of the same listener
				if (core.fn._listeners[evt].indexOf(cb) >= 0) return;

				core.fn._listeners[evt].push(cb);
			},
			
			trigger: function(evt, params) {
				
				if (!core.fn._listeners || !core.fn._listeners[evt]) return [];

				var resList = [];

				for(var i=0; i<core.fn._listeners[evt].length; i++) {

					var cb = core.fn._listeners[evt][i];

					if (typeof(cb) === 'function') {
						resList.push(cb.apply(this, params));
					}

				}

				return resList;

			},
			
			onReload: function(cb) {
				core.fn.on('reload', cb);
			},
			
			onReloading: function(cb) {
				core.fn.on('reloading', cb);
			},

			loadSettings: function() {

				var defer = $q.defer();

				$.ajax({

					type: 'POST',
					dataType: 'JSON',
					url: window.kk_shop_url,

					data: {
						"kk_dropper_settings": true,
						"kk_dropper_admin_pass": window.kk_admin_secret
					},

					success: function(res) {
						if (core.settings) delete core.settings;
						core.settings = res;
						core.latestRevision = core.settings.meta;
						defer.resolve(core.settings);
					},

					error: function(err) {
						defer.reject(err);
						message.errorHTML('Fehler beim Laden der Dropper Einstellungen!', '&bull;&nbsp;<a href="https://kreativkonzentrat.de/Wiki?page=dropper/faq" target="_blank">Troubleshooting & FAQs</a><br>&bull;&nbsp;<a href="https://kreativkonzentrat.de/Dropper?ticket" target="_blank">Support-Ticket erstellen</a>');
						message.error('Fehler beim Laden der Dropper Einstellungen!', err);
					}
				});

				return defer.promise;

			},

			reload: function() {

				var defer = $q.defer();
				
				$.ajax({

					type: 'POST',
					dataType: 'JSON',
					url: window.kk_shop_url,

					data: {
						"kk_dropper_core_data": true,
						"kk_dropper_admin_pass": window.kk_admin_secret
					},

					success: function(res) {

						// remove old properties
						for(var prop in core) {
							if (prop !== 'fn' && prop !== 'settings' && prop !== 'latestRevision') delete core[prop];
						}
			
						// set the new properties
						$.extend(true, core, res);
	
						// set admin user
						core.adminUser = window.kk_admin_user;
	
						// trigger reloading event. the callbacks will return promises so that we will wait for them
						// to finish, before we send out the final core resolve
						var beforePromises = core.fn.trigger('reloading', [core]);
	
						if (beforePromises && beforePromises.length) {

							$q.all(beforePromises).then(function() {
								// set resolved
								defer.resolve(core);
								// trigger the reload event
								core.fn.trigger('reload', [core]);
							});

						} else {

							// set resolved
							defer.resolve(core);
							// trigger the reload event
							core.fn.trigger('reload', [core]);

						}
					},

					error: function(err) {
						defer.reject(err);
						message.errorHTML('Fehler bei der Dropper Initialisierung!', '&bull;&nbsp;<a href="https://kreativkonzentrat.de/Wiki?page=dropper/faq" target="_blank">Troubleshooting & FAQs</a><br>&bull;&nbsp;<a href="https://kreativkonzentrat.de/Dropper?ticket" target="_blank">Support-Ticket erstellen</a>');
						message.error('Fehler bei der Dropper Initialisierung!', err);
					}
				});

				return defer.promise;
			},

			/**
			 * Check if we have Licencing errors and push them out
			 * to the user, to notice.
			 */
			checkLicenceErrors: function() {

				var errors = "";
		
				for (var drop in core.dropLicenses) {
					var licence = core.dropLicenses[drop];
					if (licence.isError === true) {
						errors += '<p><strong>' + drop + '</strong>: ' + licence.reason;
						errors += '</p>';
					}
				}
		
				if (errors !== "") {
					errors += '<p><a class="btn btn-small btn-danger" href="mailto:info@kreativkonzentrat.de?subject=Drops%20Lizenzfehler&body=Fehlernachricht%20hier%20pasten">Fehler melden</a></p>';
					message.errorHTML("Oups, Lizenzfehler sind aufgetreten", errors);
				}
			},
			
			/**
			 * Check for Errors in Drops that are inside the List
			 */
			checkDropErrors: function() {
				var errors = "";
				if (core.settings && core.settings.drops && core.settings.drops.length > 0) {
					for (var i=0; i<core.settings.drops.length; i++) {
						var drop = core.settings.drops[i];
						var template = drop._internal.dropId;

						if (!core.templates[template]) {
							errors += "<p><strong>" + drop._internal.name + "</strong> inaktiv: Drop '" + template + "' konnte nicht gefunden werden oder ist korrupt.</p>";
						}
						else if (!core.dropLicenses[template] || core.dropLicenses[template].isValid == false) {
							var templateName = core.templates[template].name ? core.templates[template].name : template;
							errors += "<p><strong>" + drop._internal.name + "</strong> inaktiv: Drop '" + templateName + "' hat keine gültige Lizenz.</p>";
						}
					}
				}
				 if (errors !== "") {
					message.errorHTML("Einige Drops sind inaktiv!", errors);
				}
			}

		}
	};
	
	window.dropperCore = core;
	return window.dropperCore;
	
}]);