/**
 * Delta service. Manages delta loading and saving, as well as change tracking of drops
 */
angular.module('app').service('delta', ['core', 'modal', '$timeout', '$q', function(core, modal, $timeout, $q) {
	
	// singleton pattern
	if (window.dropperDelta) {
		return window.dropperDelta;
	}

	var delta = window.dropperDelta = {

		changed: [],
		removed: [],
		eventListeners: [],

		/**
		 * Adds a listener to given events that is called when the event is being triggered
		 */
		on: function(events, cb) {

			if (!angular.isString(events)) return;

			events = events.split(',');
			events = events.map(function(event) { return event.trim(); });
			
			if (!events.length) return;
			
			events.forEach(function(event) {

				if (!delta.eventListeners[event]) delta.eventListeners[event] = [];
				var listeners = delta.eventListeners[event];
				// already listening?
				if (listeners.indexOf(cb) >= 0) return;
				listeners.push(cb);
				
			});
		},
		
		/**
		 * Notifies event listeners about an event
		 */
		notify: function(event, ...params) {

			if (!delta.eventListeners[event]) return;
			var listeners = delta.eventListeners[event];
			if (!listeners || !listeners.length) return;
			listeners.forEach(function(listener) {
				listener.apply({ event: event }, params);
			});
		},
		
		// checks if the given drop is already loaded
		isLoaded: function(dropOrGuid) {
			var drop = null;
			
			if (angular.isString(dropOrGuid)) {
				for (var i=0; i<core.settings.drops.length; i++) {
					var cmp = core.settings.drops[i];
					if (cmp && cmp._internal && cmp._internal.guid === dropOrGuid) {
						drop = cmp;
						break;
					}
				}
			} else {
				drop = dropOrGuid;
			}

			if (!drop || drop.__partial) return false;
			return true;

		},
		
		// loads a drop
		loadDrop: function(drop, reloadClean) {

			return $q(function(resolve, reject) {
				
				delta.loadDrops([drop], reloadClean).then(function(drops) {
					
					if (angular.isArray(drops) && drops.length === 1) {
						return resolve(drops[0]);
					} else {
						reject('Could not load drop, because it doesnt exist remotely anymore!');
					}
				}, function(err) {
					reject(err);
				});

			});

		},
		
		// loads a list of drops
		loadDrops: function(drops, reloadClean) {

			return $q(function(resolve, reject) {
				
				var guids = [];
				
				// figure out for which drops we need to pull in the data
				drops.forEach(function(drop) {
					if (!drop || !drop._internal || !drop._internal.guid) return;
					if (!delta.isLoaded(drop) || (reloadClean && !delta.isChanged(drop))) {
						guids.push(drop._internal.guid);
					}
				});
				
				// everything loaded already? Resolve with the input...
				if (!guids.length) return resolve(drops);

				console.log('fetching drops from remote', guids);

				// make the ajax call
				$.ajax({
					type: 'POST',
					url: core.shopURL,
					data: {
						"kk_dropper_admin_pass": core.adminSecret,
						"kk_dropper_get_drops_data": guids.join(',')
					},
					success: function(dropsData) {

						var dropsDataMap = {};

						dropsData.forEach(function(dropData) {
							if (dropData && dropData._internal && dropData._internal.guid) {
								dropsDataMap[dropData._internal.guid] = dropData;
							}
						});
						
						for (var i=0; i<drops.length; i++) {

							var drop = drops[i];
							var guid = drop._internal.guid;
							var dropData = dropsDataMap[guid] || null;

							// only do this on a match							
							if (dropData) {
								
								// backup the revision
								var dropDataRev = dropData._internal.revision;
								
								// do not copy over the remote datas internal object. We take the one that we already have
								delete(dropData._internal);
								
								// copy over remote stuff (except the revision which we restore later on)
								$.extend(true, drop, dropData);
								drop._internal.revision = dropDataRev;
	
								// this is no partial anymore
								delete(drop.__partial);
								
								// remove from the map as we have already worked on it
								delete dropsDataMap[guid];
								
							}
							// drop that should be loaded has not been found, remove it from the collection
							else {
								drops.splice(i, 1);
								i--;
							}

						}

						resolve(drops);

					},
					error: function(err) {
						reject(err);
					}
				});

			});
			
		},

		// Searches for a drop in settings
		findDrop: function(dropOrGuid) {
			if (!dropOrGuid) return null;
			var drop = core.settings.drops.find(function(drop) {
				var guid = angular.isString(dropOrGuid) ? dropOrGuid : dropOrGuid._internal.guid;
				return drop._internal.guid === guid;
			});
			return drop || null;
		},
		
		// Checks if the given drop is part of the settings
		isStoredDrop: function(dropOrGuid) {
			return !!delta.findDrop(dropOrGuid);
		},

		// gets a list of all loaded drops
		getLoaded: function() {

			var loaded = {};

			for (var i=0; i<core.settings.drops.length; i++) {
				var drop = core.settings.drops[i];
				var guid = drop._internal.guid;
				if (delta.isLoaded(guid)) {
					loaded[guid] = drop;
				}
			}
			return loaded;
		},

		// notifies that a drop was changed or resets the change state
		setChanged: function(drop, state) {

			if (!angular.isDefined(state)) state = true;

			var guid = angular.isString(drop)
							? drop
							: angular.isObject(drop)
								? drop._internal.guid
								: null;
								
			if (!guid) return;
			
			var idx = delta.changed.indexOf(guid);
			var alreadyChanged = idx >= 0;

			// set changed if not done already and this is a drop inside the drops collection			
			if (state && !alreadyChanged) {
				if (!delta.isStoredDrop(guid)) return;
				delta.changed.push(guid);
			}

			// set unchanged but only if it is changed already
			if (!state && alreadyChanged) { 
				delta.changed.splice(idx, 1);
				// notify unchange
				delta.notify('unchange', guid);
			}

			// notify listeners
			if (state) delta.notify('change', guid) ;

		},
		
		// sets all known drops to changed state
		setAllChanged: function() {
			core.settings.drops.forEach(function(drop) {
				delta.setChanged(drop);	
			});
		},
		
		// unsets changes
		resetChanged: function() {
			delta.changed = [];
		},

		// checks if a drop has been changed since app loading
		isChanged: function(drop) {

			var guid = angular.isString(drop) 
							? drop 
							: angular.isObject(drop)
								? drop._internal.guid
								: null;
			return (delta.changed.indexOf(guid) >= 0);
			
		},
		
		// gets a list of all changed drops
		getChanged: function() {

			var changed = {};

			for (var i=0; i<core.settings.drops.length; i++) {
				var drop = core.settings.drops[i];
				var guid = drop._internal.guid;
				if (delta.isChanged(guid)) {
					changed[guid] = drop;
				}
			}
			return changed;
		},

		// notifies that a drop has been removed
		setRemoved: function(drop) {
			var guid = angular.isString(drop)
							? drop
							: angular.isObject(drop)
								? drop._internal.guid
								: null;
			if (!guid) return;
			var idx = delta.removed.indexOf(guid);
			if (idx < 0) delta.removed.push(guid);

			// notify listeners
			delta.notify('remove', guid);
			
			// also remove it from the changed list
			delta.setChanged(guid, false);
		},

		// tracks all drops as removed
		setAllRemoved: function() {
			core.settings.drops.forEach(function(drop) {
				delta.setRemoved(drop);	
			});
		},
		
		 // tell us that no drops have been removed yet
		resetRemoved: function() {
			delta.removed = [];
		},
		
		// Gets a list of removed drops (guids)
		getRemoved: function() {
			return delta.removed;
		},

		// Save the settings via Ajax
		saveDeltaSettings: function(force) {
			
			if (!angular.isDefined(force)) force = false;
			
			return $q(function(resolve, reject) {

				core.settings.dropperVersion = core.dropperVersion;
				core.latestRevision = { revision: core.latestRevision.revision + 1, user: core.adminUser };
				core.settings.meta = core.latestRevision;

				// make a copy	
				var settings = JSON.parse(angular.toJson(core.settings));
				delete (settings.drops);
		
				// make a copy of the changed drops
				var changed = JSON.parse(angular.toJson(delta.getChanged()));
				
				// build the delta
				var deltaData = {
					changed: changed,
					removed: delta.getRemoved(),
					available: core.settings.drops.map(function(drop) {
						return drop._internal.guid;
					}),
					settings: settings
				};
	
				// save it!
				$.ajax({
					type: 'POST',
					url: core.shopURL,
					data: {
						"kk_dropper_save_delta": angular.toJson(deltaData),
						"kk_dropper_save_delta_force": force ? "1": "",
						"kk_dropper_save_user": core.adminUser,
						"kk_dropper_admin_pass": core.adminSecret
					},
					success : function(res) {
	
						console.log('saved: ', res);
						
						delta.resetRemoved();

						// update saved drops to recent revisions
						if (res && angular.isObject(res.revisions)) {

							for (var guid in res.revisions) {

								var drop = delta.findDrop(guid);

								if (!drop) continue;

								drop['_internal']['revision'] = res.revisions[guid];
								delta.setChanged(drop, false);

							}
						}
						
						// handle conflicts by showing the conflict modal
						if (res && res.conflicts && res.conflicts.length) {

							modal.show('SaveConflictModal', 

								// callback from modal
								function(theirs /* [...drops] */) {

									// cancelled?
									if (!angular.isArray(theirs)) {
										return resolve(res);
									}

									// replace it
									theirs.forEach(function(drop) {

										var other = delta.findDrop(drop._internal.guid);
										if (!other) return;

										// replace (without loosing reference because it might be the activeDrop)
										for (var prop in other) {
											if (prop.indexOf('$') === 0) continue;
											delete other[prop];
										}
										// copy over all props
										$.extend(true, other, drop);

										// we are already the latest revision on the server, so no change here
										delta.setChanged(other, false);

									});
									
									// save the rest of the changed drops (forced)
									if (delta.changed.length) {
										delta.saveDeltaSettings(true).then(
											// success
											function(res) {
												resolve(res);
											}, 
											// error
											function(err) {
												reject(err);
											}
										);										
									} 
									// no further changed drops (e.g. because all remote drops have been taken)
									else {
										resolve(res);
									}

									
								}, { conflicts: res.conflicts }
							);
						} 
						// no conflicts were found
						else {
							resolve(res);
						}

					},

					error: function(err) {
						reject(err);
					}
				});				
			});
		}
		
	};
	
	return window.dropperDelta;

}]);