/**
 * Window Service
 * Allows the creation of Stacked editing Windows
 */
angular.module('app').service('editWindow', ['$compile', '$rootScope', '$controller', '$timeout', 'state', 'core', 'bootstrap', function($compile, $rootScope, $controller, $timeout, state, core, bootstrap) {

	// global
	if (!state.editWindows) state.editWindows = [];
	
	var self = this;

	// list with all open windows (scopes)
	this.windows = state.editWindows;

	// closes all windows
	this.closeAll = function(cb) {

		for(var i=0; i<self.windows.length; i++) {

			var wnd = self.windows[i];

			wnd.close(function() {
				if (self.windows.length == 0 && typeof(cb) == 'function') {
					cb();
				}
			}, true);

		}	

	};

	// shows a window
	this.show = function(nest, options, injects) {

		var self = this;

		// delay the showing of a window so that other windows can close first
		$timeout(function() {

			var nested = nest ? true : false;

			var windowScope = $rootScope.$new(true);

			windowScope.nested = nested;
			
			if (angular.isString(options)) {
				options = {
					template: core.adminURL + 'js/views/windows/' + options.toLowerCase() + '.html.js?' + core.cacheID,
					controller: options + "Controller"
				};
			}

			windowScope.options = options;

			// called when the template inside the window is ready and loaded
			windowScope._loadListeners = [];

			windowScope._load = function(cb) {

				windowScope._loaded ? cb() : windowScope._loadListeners.push(cb);

			};
			
			// when content is loaded call the callbacks
			windowScope.$on('$includeContentLoaded', function () {

				if (windowScope._loaded) return;
				windowScope._loaded = true;

				$timeout(function() {

					for(var i=0; i<windowScope._loadListeners.length; i++) {
						windowScope._loadListeners[i]();
					}

					windowScope._loadListeners = [];

				});

			});

			// build the window template
			var template = '<div class="hero-unit edit-window" style="opacity: 0" ng-class="{\'actives\':isActive}" ng-click="activate($event)">\
								<ng-include src="options.template" class="edit-window-content"></ng-include>\
							</div>';

			windowScope.dom = $compile(template)(windowScope);

			// make us known to the collection
			windowScope.nested ? self.windows.push(windowScope) : self.windows.unshift(windowScope);

			// create the controller, if any given
			if (options.controller) {

				var ctrlInjects = {
					"$windowScope": windowScope, 
					"$element": windowScope.dom, 
					"useConditional": true,
					"nameBinding": null
				};

				if (angular.isObject(injects)) {
					ctrlInjects = angular.extend({}, ctrlInjects, injects);
				}

				$controller(options.controller, ctrlInjects);

			}

			// closes the window
			windowScope.close = function(cb, fastClose) {

				var fast = !windowScope.nested || fastClose ? true : false;
				var animDuration = fast ? 300 : 400;

				if (fast) {
					windowScope.dom.addClass('fast');
				}

				// animate out
				windowScope.dom.css({
					opacity: 0,
					transform: 'translateY(250px)'
				});
				
				$timeout(function() {
					
					// remove from the dom
					windowScope.dom.remove();

					// destroy our scope
					windowScope.$destroy();

					// remove ourself from the windows list
					var idx = self.windows.indexOf(windowScope);

					// remove from the window list
					if (idx >= 0) self.windows.splice(idx, 1);

					// get the parent, if exists
					var parent = idx > 0 ? self.windows[idx-1] : null;

					// activate parent?
					if (parent) {
						parent.activate();
						parent.$apply();
					}

					if (typeof(cb) === 'function') cb(idx);

				}, animDuration+100);

			};

			// sets this window active
			windowScope.activate = function(e) {

				if (windowScope.isActive) return;

				// close all windows higher than ourself
				var wndIdx = self.windows.indexOf(windowScope);

				if (wndIdx >= 0 && self.windows.length > wndIdx) {
					for(var i=wndIdx+1; i<self.windows.length; i++) {
						self.windows[i].close();
					}
				}

				// make all windows inactive
				for(var j=0; j<self.windows.length; j++) {
					self.windows[j].isActive = false;
				}

				windowScope.isActive = true;

			};

			// shows the window
			windowScope._show = function() {

				windowScope._load(function() {

					// get the window below this one
					var windowIndex = self.windows.indexOf(windowScope);
					var marginTop = 0;
	
					if (windowIndex > 0) {

						// find a scroll container
						var container = $('#edit-windows');
						
						while (container.css('overflow-y') !== 'auto' && container.css('overflow-y') !== 'scroll' && !container.is(document)) {
							container = container.parent();
						}

						marginTop = Math.max(container.scrollTop(), windowIndex * 80);

					}

					// append us to the dom
					$('#edit-windows').append(windowScope.dom);
	
					windowScope.activate();
	
					// set the width
					windowScope.dom.css('width', (720 + 30*self.windows.length - 30/self.windows.length) + 'px');
	
					var fast = !windowScope.nested;
	
					// wait a short time before animating, because we want the dom to settle
					$timeout(function() {
						
						if (fast) windowScope.dom.addClass('fast');

						windowScope.dom.addClass('actives').css({
							'margin-top': marginTop + 'px',
							'transform': 'unset',
							'opacity': 1
						});

					}, 50);
					
					bootstrap.rebind();							
				});

			};

			//
			// SHOW US
			//
			if (nested || self.windows.length === 1) {

				// show ourself when nested
				windowScope._show();

			} else {

				// close all other windows if not nested and wait for the last one to close
				for(var i=0; i<self.windows.length; i++) {

					var wnd = self.windows[i];

					if (wnd !== windowScope) {
						wnd.close(function() {
							if (self.windows.length === 1) {
								windowScope._show();
							}
						}, true);
					}

				}
			}
			
		}, 10);
	};

}]);