/**
 * State service. Holds data belonging to the local application state
 */
angular.module('app').service('state', ['core', 'upgrade', 'packageManager', 'message', 'modal', 'delta', '$timeout', '$q', function(core, upgrade, packageManager, message, modal, delta, $timeout, $q) {
	
	// singleton pattern
	if (window.dropperState) {
		return window.dropperState;
	}

	var SIDEBAR_DEFAULT_WIDTH = 360;		
	
	var state = window.dropperState = {

		// currently visible drop that is being edited
		activeDrop: null,

		// reference to a drop that is being opened but still loading
		activeDropLoading: null,
				
		// indicates if there were changes to the settings object since the last save 
		dirty: false,
		
		// indicates if a save is in progress
		saveInProgress: false,
		
		// indicates if the dashboard is shown to the user
		dashboardOpen: true,

		// holds state of the sidebar
		sidebar: {
			get width() {
				var width = +window.localStorage.getItem('kk_dropper.sidebarWidth') || SIDEBAR_DEFAULT_WIDTH;
				return width;
			},
			set width(val) {
				var width = angular.isNumber(val) ? +val : SIDEBAR_DEFAULT_WIDTH;
				window.localStorage.setItem('kk_dropper.sidebarWidth', width);
			}
		},

		// holds history state and actions
		history: {

			// saves the history of open drops
			items: [],
			
			// current active history item
			current: null,
			
			// pushes a new item on the history
			push: function(drop) {

				state.history.cleanup();
	
				// no drop given?
				if (!angular.isObject(drop)) return;
	
				// not available in our common drop list?
				if (core.settings.drops.indexOf(drop) < 0) return;

				// if this drop is in history already, remove it
				var idx = state.history.items.indexOf(drop);
				if (idx >= 0) state.history.items.splice(idx, 1);

				// remove all items that follow the history.current drop
				var currentIdx = state.history.items.indexOf(state.history.current);
				state.history.items.splice(currentIdx+1);
	
				// push us on top of the history
				state.history.items.push(drop);
				state.history.current = drop;
				
			},
			
			// gets the previous history item
			back: function(getOnly) {

				state.history.cleanup();

				var idx = state.history.items.indexOf(state.history.current);
				if (idx <= 0) return null;
				
				var prev = state.history.items[idx-1];
				if (getOnly) return prev;
				
				state.history.current = prev;

				return state.history.current;
	
			},
			
			// gets the next history item
			forward: function(getOnly) {

				state.history.cleanup();
	
				var idx = state.history.items.indexOf(state.history.current);
				if (idx >= state.history.items.length-1) return null;
				
				var next = state.history.items[idx+1];

				if (getOnly) return next;
				
				state.history.current = next;

				return state.history.current;
	
			},

			// checks if back navigation is possible
			backPossible: function() {
				state.history.cleanup();
				var idx = state.history.items.indexOf(state.history.current);
				return (idx > 0);
			},
			
			// checks if forward navigation is possible
			forwardPossible: function() {
				state.history.cleanup();
				var idx = state.history.items.indexOf(state.history.current);
				return (idx >= 0 && idx < state.history.items.length-1);
			},
			
			// removes history items, that are not inside our settings anymore
			cleanup: function() {
				for (var i=0; i<state.history.items.length; i++) {
					let historyItem = state.history.items[i];
					if (core.settings.drops.indexOf(historyItem) >= 0) continue;
					state.history.items.splice(i, 1);
					i--;
				}
			}
		},

		/**
		 * set the current working state diry or clean
		 */
		setDirtySaveState: function(isDirty) {
			state.dirty = !!isDirty;
		},

		/**
		 * Shows the dashboard to the user
		 */
		showDashboard: function() {
			state.dashboardOpen = true;
		},
		
		/**
		 * Hides the dashboard from the user
		 */
		hideDashboard: function() {
			state.dashboardOpen = false;
		},
		
		/**
		 * Toggles the visibility state of the dashboard
		 */
		toggleDashboard: function() {
			state.dashboardOpen = !state.dashboardOpen;
		},

		/**
		 * Verifiy data integrity
		 */
		verifyChanges: function() {
			
			console.log('verfiying changes and running updates');

			// run upgrade manager
			state.runUpgrades(function() {
				
				// checkfor multilang errors
				upgrade.updateMultiLangFields(core.settings);
	
				// check licence errors
				core.fn.checkLicenceErrors();
	
				// check for invalid drops
				core.fn.checkDropErrors();
				
				// trigger verify changes event
				core.fn.trigger('verifyChanges');
				
			});

			// show dashboard
			state.showDashboard();

			//  check for forced upgrades
			packageManager.checkForcedUpgrades();
			
		},
		
		/**
		 * Runs Dropper and Drop upgrades on the core settings.
		 * Uses the message service to notify about success and errors
		 */
		runUpgrades: function(cb) {
			
			upgrade.runUpgrades(core.settings, function(upgradeInfos) {

				if (angular.isFunction(cb)) cb(upgradeInfos);

				if (upgradeInfos.errors.length) {
	
					for (var i=0; i<upgradeInfos.errors.length; i++) {
						var error = upgradeInfos.errors[i];
						message.error('Fehler beim ' + (error.type === 'dropper' ? 'Dropper' : 'Drop') + ' Upgrade', error.message);
					}
	
					return;
				}
	
				if (upgradeInfos.settings) {
	
					core.settings = upgradeInfos.settings;
					state.setDirtySaveState(true);
					
					if (upgradeInfos.dropperUpgraded) {
	
						message.success('Dropper Upgrade auf v' + upgradeInfos.settings.dropperVersion + ' erfolgreich abgeschlossen!', 'Bitte speichern um die Änderungen zu übernehmen!', 0);
						state.setDirtySaveState(true);
	
					}
					
					if (upgradeInfos.dropUpgrades.length) {
	
						state.setDirtySaveState(true);
						
						for (var i=0; i<upgradeInfos.dropUpgrades.length; i++) {
							
							var dropInfo = upgradeInfos.dropUpgrades[i];
							message.successHTML('<strong>"' + (dropInfo.template && dropInfo.template.name ? dropInfo.template.name : '') + '"</strong> Drops wurden auf Version ' + dropInfo.template.version + ' aktualisiert.<br>Bitte speichern um die Änderungen zu übernehmen.', '<strong>' + dropInfo.drop._internal.name + '</strong> wurde erfolgreich aktualisiert', 0);
							
						}
	
					}
				}				
			});

		},

		/**
		 * Loads the settings
		 */
		loadSettings: function() {
			return core.fn.loadSettings();
		},

		/**
		 * Save the Settings via Ajax
		 */
		saveSettings: function(rev) {
			
			state.saveInProgress = true;

			return $q(function(resolve, reject) {
				
				var beforeSavePromises = core.fn.trigger('beforeSave', []);

				// await all before promises
				$q.all(beforeSavePromises).then(function() {

					delta.saveDeltaSettings(rev).then(
						
						// success
						function(res) {
							$timeout(function() {
								state.saveInProgress = false;
								state.setDirtySaveState(delta.changed.length ? true : false);
							});
							core.fn.trigger('saveSettings', [null, res]);
							resolve(res);
						},
						
						// error
						function(err) {
							$timeout(function() {
								state.saveInProgress = false;
								message.error('Beim Speichern ist ein Fehler aufgetreten: ', angular.fromJson(err).statusText || err);
							});
							core.fn.trigger('saveSettings', [err, null]);
							reject(err);
						}
					);
				}).catch(function(err) {
					message.error('Vor dem Speichern ist ein Fehler aufgetreten:', err);
				});

			});
		}

	};

	// listen for drop changes to set the dirty save state
	delta.on('change, unchange, remove', function(guid) {
		if (delta.changed.length || delta.removed.length) state.setDirtySaveState(true);
	});
	
	return window.dropperState;

}]);