/**
 * Upgrade Service that is able to upgrade Drop Data.
 * There are two different scenarios: Dropper Upgrades and
 * Drop upgrades form the upgrade.js file
 */
angular.module('app').service('upgrade', ['message', 'core', 'delta', 'environment', '$q', function(message, core, delta, environment, $q) {

	if (window.kkUpgrader) return window.kkUpgrader;

	var upgrader = window.kkUpgrader = {

		_dropperUpgrades: [],
		_dropUpgrades: [],

		/**
		 * Run Upgrades on the given settings data
		 */
		runUpgrades: function(settings, cb /* upgradeState */) {
	
			var upgradeState = {

				dropperUpgraded: false,
				dropUpgrades: [],
				errors: [],
				settings: null
				
			};
			
			// run control center upgrades
			upgrader.runDropperUpgrades(settings, function(upgraded, count) {
	
				// error?
				if (angular.isString(upgraded)) {

					upgradeState.errors.push({
						type: 'dropper',
						message: upgraded
					});

					return cb(upgradeState);

				}
	
				// successfull upgrade?
				if (count > 0 && upgraded) {

					upgradeState.settings = upgraded;
					upgradeState.dropperUpgraded = true;
					
					// mark all drops in this set as changed
					upgraded.drops.forEach(function(drop) {
						delta.setChanged(drop);
					});

				}
	
				// run drop upgrades
				upgrader.runDropUpgrades(upgraded, function(upgraded, upgradeInfo) {

					if (upgradeInfo && upgradeInfo.length) {

						for (var i=0; i<upgradeInfo.length; i++) {

							var info = upgradeInfo[i];

							if (info.error) {
								
								upgradeState.errors.push({
									type: 'drop',
									message: info.error
								});
								
							}
							else {
								upgradeState.dropUpgrades.push(info);
							}

						}

						upgradeState.settings = upgraded;

					}
					
					cb(upgradeState);

				});

			});

		},

		/**
		 * Unregisters all upgrades
		 */
		cleanup: function() {
			upgrader._dropUpgrades = [];
			upgrader._dropperUpgrades = [];
		},

		/**
		 * Adds a new upgrade for the control center.
		 * @from - the version from which to upgrade (e.g. "100.1")
		 * @to - the version to which to upgrade (e.g. "100.2")
		 * @upgradeFn - the upgrade function, params: settings, templates
		 */
		addDropperUpgrade: function(from, to, upgradeFn) {

			console.log('adding dropper upgrade', arguments);

			upgrader._dropperUpgrades.push({
				from: from,
				to: to,
				fn: upgradeFn
			});

		},

		/**
		 * Recursively runs all upgrades on the given data
		 * @settings - the Dropper settings object
		 * @returns - the upgraded settings object
		 */
		runDropperUpgrades: function(settings, cb /* changedSettings, count */) {

			if (!angular.isFunction(cb)) return console.log('Attempting to run Dropper Upgrades, but no callback given!');

			console.log('running dropper upgrades', upgrader);

			// check if there are upgrades available
			var upgradesAvailable = false;
			upgrader._dropperUpgrades.forEach(function(upgrade) {
				if (upgrade.from === settings.dropperVersion) {
					upgradesAvailable = true;
				}
			});

			if (!upgradesAvailable) return cb(settings, 0);
			upgrader._runDropperUpgrades(settings, 0, cb);

		},

		/**
		 * Recursively runs all upgrades on the given data
		 * @settings - the Dropper settings object
		 * @count - the Count of upgrades that where run
		 * @returns - the upgraded settings object
		 */
		_runDropperUpgrades: function(settings, count, cb) {

			// make sure all drop data is loaded before performing an upgrade
			delta.loadDrops(settings.drops || [], false).then(

				// success
				function() {
	
					var clonedSettings = $.extend(true, {}, settings);
					
					if (!count) {
						count = 0;
					}
		
					if (!clonedSettings.dropperVersion) {
						clonedSettings.dropperVersion = "100.0";
					}
		
					for(var i=0; i<upgrader._dropperUpgrades.length; i++) {
						var upgrade = upgrader._dropperUpgrades[i];
						
						if (clonedSettings.dropperVersion === upgrade.from) {
							// count one up
							count ++;
							console.log("Upgrading Dropper", upgrade.from, upgrade.to);
							var up = upgrade.fn(clonedSettings, core.templates);
		
							// recursively upgrade further
							if (!up || typeof(up) === 'string') {
								// error at last upgrade
								if (!up) up = 'Das Migrationsresultat war leer.';
								return cb('Das Dropper Upgrade von ' + upgrade.from + ' auf ' + upgrade.to + ' ist fehlgeschlagen: ' + up, count);
							}
		
							up.dropperVersion = upgrade.to;
							return upgrader._runDropperUpgrades(up, count, cb);
						}
					}
		
					return cb(settings, count);				
				},
				
				// error
				function(err) {
					return cb('Ein oder mehrere Drops konnte nicht geladen werden. ' + err, count);
				}
			);

		},

		/**
		 * Adds a new upgrade for an drop.
		 * @dropId - the ID of the drop
		 * @from - the version from which to upgrade (e.g. "1.0.0")
		 * @to - the version to which to upgrade (e.g. "1.0.1")
		 * @upgradeFn - the upgrade function, params: drop, template
		 */
		addDropUpgrade: function(dropId, from, to, upgradeFn) {

			console.log('adding drop upgrade', arguments);

			upgrader._dropUpgrades.push({
				dropId: dropId,
				from: from,
				to: to,
				fn: upgradeFn
			});

		},

		/**
		 * Returns a list of drops for which we have got upgrades available
		 */
		_getUpgradeableDrops: function(settings) {
			
			var drops = [];

			settings.drops.forEach(function(drop) {

				var version = drop._internal.version || "1.0.0";
				
				upgrader._dropUpgrades.forEach(function(upgrade) {
					if (drop._internal.dropId === upgrade.dropId && version === upgrade.from) {
						if (drops.indexOf(drop) < 0) drops.push(drop);
					}
				});
			});
			
			return drops;
		},

		/**
		 * Recursively runs all upgrades on the given data
		 * @settings - the Dropper settings object
		 * @returns - the upgraded settings object
		 */
		runDropUpgrades: function(settings, cb /* upgradedSettings, count */) {

			console.log('running drop upgrades', this);

			// precheck which drops can be upgraded
			var dropsToUpgrade = upgrader._getUpgradeableDrops(settings);
			
			// make sure they are loaded
			delta.loadDrops(dropsToUpgrade).then(

				// all drops loaded
				function() {
	
					var upgradeState = [];
		
					for(var i=0; i<settings.drops.length; i++) {
		
						var drop = settings.drops[i], template;
		
						if (core.templates[drop._internal.dropId]) {
							template = core.templates[drop._internal.dropId];
						} else {
							continue;
						}
		
						// run upgrade on this drop
						upgrader._runDropUpgrades(drop, template, 0, $.proxy(function(upgradedDrop, count) {
		
							// nothing upgraded?
							if (count === 0) return;
		
							// upgrade success?
							if (upgradedDrop && !angular.isString(upgradedDrop)) {
								settings.drops[this.index] = upgradedDrop;
								upgradeState.push({
									drop: drop,
									error: false,
									count: count,
									template: template
								});
							} 
							// error
							else {
								upgradeState.push({
									drop: drop,
									error: upgradedDrop,
									count: count,
									template: template
								});
							}
		
						}, { index: i }));
					}
		
					return cb(settings, upgradeState);				
				},

				function(err) {
					message.error('Upgrade nicht erfolgreich!', err);
					return cb(settings, null);
				}
			);

		},


		/**
		 * Recursively runs all upgrades on the given data
		 * @drop - the drop data to upgrade
		 * @count - a number of upgrades already run on the data
		 * @returns - the upgraded settings object
		 */
		_runDropUpgrades: function(drop, template, count, cb) {

			var clonedDrop = $.extend(true, {}, drop);

			if (!count) {
				count = 0;
			}

			if (!clonedDrop._internal.version) {
				clonedDrop._internal.version = "1.0.0";
			}

			for(var i=0; i<upgrader._dropUpgrades.length; i++) {

				var upgrade = upgrader._dropUpgrades[i];
				
				if (clonedDrop._internal.dropId === upgrade.dropId && clonedDrop._internal.version === upgrade.from) {
					// count one up
					count ++;
					var up = upgrade.fn(clonedDrop, template);
					console.log("Drop Upgraded: ", upgrade.from, upgrade.to, up);
					

					// recursively upgrade further
					if (!up || typeof(up) === 'string') {
						// error at last upgrade
						if (!up) up = "Das Migrationsresultat war leer.";
						return cb('Das Upgrade des Drops "' + drop._internal.name + '" von ' + upgrade.from + ' auf ' + upgrade.to + ' ist fehlgeschlagen: ' + up, count);
					}

					up._internal.version = upgrade.to;

					// track upgrade change
					delta.setChanged(up);

					// call us again, we might have further updates
					return upgrader._runDropUpgrades(up, template, count, cb);
				}
			}

			return cb(drop, count);

		},

		/**
		 * Check if there is a drop option defined for the given template ID
		 */
		dropHasOption: function(templateId, optionName) {

			var templates = core.templates;

			if (typeof templates === 'object' && typeof templates[templateId] === 'object') {

				var template = templates[templateId];

				if (template.config instanceof Array) {
					for (var i=0; i<template.config.length; i++) {
						if (template.config[i].property === optionName) return true;
					}
				}

				if (template.settings instanceof Array) {
					for (var i=0; i<template.settings.length; i++) {
						if (template.settings[i].property === optionName) return true;
					}
				}
			}

			return false;

		},

		/**
		 * Check if there is a dropItem option defined for the given template
		 */
		dropItemHasOption: function(templateId, optionName) {

			var templates = core.templates;

			if (typeof templates === 'object' && typeof templates[templateId] === 'object') {
				var template = templates[templateId];
				if (template.itemSettings instanceof Array) {
					for (var i=0; i<template.itemSettings.length; i++) {
						if (template.itemSettings[i].property == optionName) {
							return true;
						}
					}
				}
			}

			return false;

		},

		/**
		 * Iterates over a string type object and calls the callback function which in turn can edit
		 * split out text parts. This is very usefull for multilanguage fields.
		 */
		editText: function(obj, fn /* text, iso */) {

			// simple string
			if (angular.isString(obj)) {
				return fn(obj, environment.languageDefault.cISO);
			}

			if (angular.isObject(obj)) {

				// mutlilang object?
				if (obj['_multilanguage']) {
					var cloned = $.extend(true, {}, obj);
					var langCounter = 0;
					for(var lang in obj) {
						if (lang !== '_multilanguage') {
							cloned[lang] = fn(cloned[lang], lang);
							langCounter ++;
						}
					}
					// maybe the object looked like {_multilanguage: true} but no content set
					// handle it
					if (langCounter === 0) {
						cloned[environment.languageDefault.cISO] = fn("", environment.languageDefault.cISO);
					}
					return cloned;
				}
				// no multilang object
				else {
					return obj;
				}

			}

			// not set yet... but maybe you want to?
			if (typeof(obj) === "undefined" || obj === null) {
				return fn(null, environment.languageDefault.cISO);
			}

			// default
			return obj;

		},

		urlToLink: function(url, title) {

			return upgrader.editText(url, function(url, iso) {

				return {
					_url: true,
					value: url,
					type: "Manual",
					title: (typeof(title) === "string" ? title : (angular.isObject(title) && angular.isString(title[iso]) ? title[iso] : ""))
				};

			});		

		},

		/**
		 * Searches for drops that have a plain string given but should be multilang now
		 */
		updateMultiLangFields: function(settings) {

			var changed = 0;
			var changedArr = [];
	
			for (var dropID in settings.drops) {
				var drop = settings.drops[dropID];
				if (!drop._internal || !drop._internal.dropId) {
					console.log("Could not update Multilang Fields because no _internal or dropId set", drop);
					continue;
				}
	
				var dropSetting = core.templates[drop._internal.dropId];
				if (dropSetting && dropSetting.config) changedArr.concat(upgrader._updateMultiLangFields(drop, dropSetting.config));
				if (dropSetting && dropSetting.settings) changedArr.concat(upgrader._updateMultiLangFields(drop, dropSetting.settings));
			}
			
			if (changedArr.length > 0) {
				message.info(changed + ' Objekte wurden mit Mehrsprachigkeit nachgerüstet.', 'Bitte Speichern um Änderungen zu übernehmen.');
				console.log('multilang update: ', changedArr, false);
			}

		},

		/**
		 * Recursively loops over the config properties and looks in the obj if we have got
		 * string properties that should be multilang.
		 */
		_updateMultiLangFields: function(obj, cfg) {

			var changedArr = [];
	
			for (var i=0; i<cfg.length; i++) {

				var option = cfg[i];

				if (angular.isString(obj[option.property]) && (option.multilanguage === true || option.multilanguage === "true")) {

					console.log("changing drop property: ", obj, "from value:", obj[option.property], "at property:", option.property, "to multilanguage");

					changedArr.push({ item: obj, prop: option.property });

					var oldVal = obj[option.property];
					obj[option.property] = { _multilanguage: true };
					obj[option.property][environment.languageDefault.cISO] = oldVal;

				}

				if (angular.isArray(obj[option.property]) && option.type === 'list') {

					changedArr.concat(upgrader._updateMultiLangFields(obj[option.property], option.itemSettings));

				}

			}

			return changedArr;

		}
		
	};

	/**
	 * This will return the upgrader
	 */
	return window.kkUpgrader;
	
}]);