angular.module('app').controller('GroupImproveModalController', ['$modalScope', 'core', 'utils', 'delta', function($modalScope, core, utils, delta) {

	$modalScope.renames = {};

	// tries to normalize some common naming conventions in
	// a name that might be there to make a cluster from it
	var normalizeName = function(name) {
		return name.replace(/([A-Za-z])-([A-Za-z])/g, '$1 $2');
	};

	// Gets a list of clusters from a list of names
	var getClusters = function(names) {

		let clusters = {};
	
		for (let name of names) {
			
			// remove the last part of a grouping because this is most likely the drop name
			let norm = normalizeName(name).split(':');
			norm.pop();
			norm = norm.join(':');
	
			let parts = norm.split(' ');
			let comb = '';
	
			for (let part of parts) {
				comb = (comb + ' ' + part).trim();
				let combPath = comb.toLowerCase();
				if (!clusters[combPath]) { 
					clusters[combPath] = {
						path: combPath,
						name: comb,
						count: 1
					};
				} else {
					clusters[combPath].count++;
				}
			}
		}
	
		clusters = Object.values(clusters);
		clusters = clusters.filter(function(cluster) { return cluster.count > 1; });
		clusters = clusters.sort(function(a, b) { return b.name.length - a.name.length});
	
		// we have to cleanup clusters that start with the name of another cluster, but contain exactly the same count.
		// otherwise we would create unnecccary grouping
		for (let clusterA of clusters) {
			for (let clusterB of clusters) {
				if (clusterA === clusterB) continue;
				if (clusterB.count === clusterA.count && clusterA.path.indexOf(clusterB.path) === 0) {
					clusterB.ignore = true;
				}
			}
		}
	
		clusters = clusters.filter(function(cluster) { return !cluster.ignore});
	
		return clusters;
	};
	
	// optimizes a given name based on the clusters
	var optimizeName = function(name, clusters) {

		name = normalizeName(name);
		
		for (let cluster of clusters) {
			if (name.toLowerCase().indexOf(cluster.name.toLowerCase() + ' ') !== 0) continue;
			name = cluster.name + ': ' + name.substring(cluster.name.length);
			name = name.replace(/\ {2,}/gm, ' ');
			name = name.replace(/:{2,}/gm, ':');
		}

		return name;

	};

	/**
	 * Tries to group all Drop names based on name clusters.
	 */
	var autoGroupDrops = function() {

		let changes = [];
		
		// get new names for all known drops
		let names = core.settings.drops.map(function(drop) { return drop._internal.name; });
		let clusters = getClusters(names);

		for (let drop of core.settings.drops) {
			let name = drop._internal.name;
			let newName = optimizeName(name, clusters);
			if (newName !== name) { 
				changes.push({
					oldName: name,
					newName: newName,
					drop: drop
				});
			}
		}
		
		var renames = {};
		
		// go through all changes and see what groups will be renamed
		changes.forEach(function(change) {
			var oldGroup = utils.Group.pathFromName(change.oldName);
			var newGroup = utils.Group.pathFromName(change.newName);
			if (oldGroup === newGroup) return;
			if (!renames[oldGroup]) { 
				renames[oldGroup] = {
					oldGroup: oldGroup,
					newGroup: newGroup,
					drops: [],
					take: true
				};
			}
			renames[oldGroup].drops.push({
				drop: change.drop,
				oldName: change.oldName,
				newName: change.newName
			});
		});
		
		renames = Object.values(renames);

		// sort by names
		renames = renames.sort(function(a, b) {
			if (a.newGroup.indexOf(b.newGroup) === 0) return -1;
			if (b.newGroup.indexOf(a.newGroup) === 0) return 1;
			if (a.newGroup < b.newGroup) return -1;
			if (b.newGroup < a.newGroup) return 1;
			return 0;
		});
		
		$modalScope.renames = renames;
	};

	/**
	 * Returns how many groups will be renamed
	 */
	$modalScope.getRenameCount = function() {
		var renameGroups = $modalScope.renames.filter(function(rename) {
			return !!rename.take;
		});
		return renameGroups.length;
	};

	/**
	 * Apply group rename operations
	 */
	$modalScope.applyOperations = function() {

		var renameGroups = $modalScope.renames.filter(function(rename) {
			return !!rename.take;
		});
		
		renameGroups.forEach(function(group) {
			group.drops.forEach(function(dropChange) {
				dropChange.drop._internal.name = dropChange.newName;
				delta.setChanged(dropChange.drop);
			});
		});
		
		$modalScope.close(true);
	};
	
	// group em
	autoGroupDrops();

}]);